#!/usr/bin/env python
# coding: utf-8

# # Heat Pump Example
# 
# 
# How to calculate a heat pump with carbatpy.
# There are two storages (cold and hot) running between prescribed temperatures with defined fluids.

# In[17]:


import carbatpy as cb


# ## Select the input file (yaml or json works):

# In[18]:


DEFAULT_DIR =  cb.CB_DEFAULTS["General"]["CB_DIR"]+"\\data\\"
DEFAULT_FILE = DEFAULT_DIR+"hp-input-dictvariables"


# The content of the (here: json) file is shown below.
# It includes the wanted minimum separation between fluid temperatures D_T_MIN, superheating before compressor, isentropic efficiency (=const. at the moment), the pressures, fluids and their composition in both storages and the working fluid. If **P_Working["setting"] =="auto"** , the pressures will be determined acording to the storage temperatures, but this can fail.
# Alternatively "setting" can be anything else, then the two pressures (p_low for the evaporator, p_high for the condenser, all in *Pa*) will be used together with the temperatures (and T-differences) listed. Again, if the values are not chosen properly, this will fail, therefore check the *warning* (see below).
# 
# 
# 
# 
#     {
#         "D_T_MIN": 4.0,
#         "D_T_SUPER": 5,
#         "ETA_S_C": 0.7,
#         "P_IN_STORAGE_COLD": 500000.0,
#         "P_IN_STORAGE_HOT": 500000.0,
#         "P_WORKING": {
#             "optimize": "None",
#             "setting": "auto",
#             "p_low": 0.0,
#             "p_high": 0.0
#         },
#         "Q_DOT_MIN": 1000.0,
#         "T_IN_STORAGE_COLD": 288.15,
#         "T_IN_STORAGE_HOT": 288.15,
#         "T_OUT_STORAGE_COLD": 250.15,
#         "T_OUT_STORAGE_HOT": 363.0,
#         "fluids_all": {
#             "WORKING": [
#                 "Propane * Butane * Pentane * Hexane",
#                 [
#                     0.75,
#                     0.05,
#                     0.15,
#                     0.05
#                 ]
#             ],
#             "STORAGE_HOT": [
#                 "Water",
#                 [
#                     1.0
#                 ]
#             ],
#             "STORAGE_COLD": [
#                 "Methanol",
#                 [
#                     1.0
#                 ]
#             ]
#         }
#     }
# 
# ## Read the inputs from the file and create an input dictionary:

# In[19]:


inputs = cb.hp_simple.HpVal.load_from_file(DEFAULT_FILE+".json")
# print(DEFAULT_FILE)

INPUTS = inputs.to_dict()


# ## Create a new instance of HeatPump with the wanted data:

# In[20]:


hpn = cb.hp_simple.HeatPump( INPUTS)
print(hpn.evaluation)


# **Calculate the cop** and further values of the heat pump and plot it:

# In[21]:


cop_n = hpn.calc_heat_pump(verbose=True)
hpn.hp_plot()
print(f"COP: {cop_n:.3f}")


# ## Get the results
# Further **Evaluation results** are found here:

# In[22]:


hpn.evaluation


# Mass flow rates are calculated for a fixed value of the heat flow rate at high temperature $\dot Q_h$.
# The mean temperature differences are calculted for each heat exchanger using the (typically) 50 points along the heat exchanger. It can be used to estimate heat exchanger sizes:
# $\dot Q = UA \Delta T_{mean}$
# 
# ## Check for warnings
# If the temperature curves cross with the given T-p-values, you will get warnings and you should not use the results. *warning* is a list with the problem of each device listed, if there are some:

# In[23]:


if hpn.warning:
    print(f"Problems: {hpn.warning}")


# Information about **all the calculated states** (typically 50 points per heat exchanger)
# and the modelled **components**/devices can also be obtained: 
# (They are not printed here, because the output is quite long.)

# In[24]:


all_plotted_states = hpn.all_states
component_info = hpn.components


# ## Storing
# Finally you can store all these things (in part this is done automatically, check your Results folder (if it is not set as environment variable, they are in your TEMP folder):

# In[25]:


hpn.save_to_file(cb.CB_DEFAULTS["General"]["RES_DIR"]+"\\my_results.json")
RESULTS_DIR = cb.CB_DEFAULTS["General"]["RES_DIR"]

check_res_folder = False  # set this to True, if you want check yours

if check_res_folder:
    import os
    try: # results directory set, Environment variable?
        print (os.environ['CARBATPY_RES_DIR'])
    except:
        pass
    print("RESULTS folder :", cb.CB_DEFAULTS["General"]["RES_DIR"])


# The automatically stored results can be read to a dictionary:

# In[26]:


my_dict = cb.hp_simple.read_hp_results(RESULTS_DIR +
                              "\\last_T_H_dot_plot_hp_evaluation_dict.npy")


# In[ ]:







# -*- coding: utf-8 -*-
"""

Helper functions for optimizations of cycles, which use comp.py.

Created on Mon Aug  4 11:39:05 2025

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""
from copy import deepcopy
import multiprocessing
import gc
import numpy as np
from scipy.optimize import minimize, differential_evolution, basinhopping, shgo
import carbatpy as cb


def optimize_wf_heat_pump(dir_name, config_0, bounds_0, **kwargs):
    """
    Heat pump COP optimization, varying p_high and mixture composition

    p_low cannot be selected directly, it is determined by the storage
    temperature and the superheating temperature difference and the minimum
    approach temperature. In optimization, the low pressure (wanted) from
    the conciguration yaml file is checked. If the calculated low pressure is
    below that value, the COP is weighted with the ratio.

    Parameters
    ----------
    dir_name : str or dict
        either the path to the yaml file with all configuration parameters, or
        a dictionary with all values.
    config_0 : dict
        vales set here and deviating from dir_name.
    bounds_0 : dictTYPE
        dictionary with the p_high bounds and the upper bounds of the first
        three molefractions. The remaining value is the difference to 1. It is
        checked that the sum is 1 and all vales are positive.

    **kwargs :
        - 'optimize_global': select optimization algorith "dif_evol", "bas_hop",
        'shgo', or local minimizer (Nelson Mead).

    Returns
    -------
    result : OptimizeResult
        The result of the optimization.
    paths: list of lists
        from function extract_optim_data_with_paths. Needed to reconstruct the
        config_0 dictionary from the optimizer x-list.

    """

    x0, bnds, paths = extract_optim_data_with_paths(config_0, bounds_0)
    opt_global = kwargs.get('optimize_global', "dif_evol")
    max_iter = kwargs.get('maxiter', 200)
    workers = kwargs.get('workers', 1)
    opt_options =kwargs.get('opt_options',{})
    tolerance = 1e-4

    if opt_global == "dif_evol":
        result = differential_evolution(_opti_hp_func,
                                        bnds,
                                        args=(config_0,
                                              dir_name, paths),
                                        workers=workers,
                                        maxiter=max_iter,
                                        )
    elif opt_global == "bas_hop":
        result = basinhopping(_opti_hp_func,
                              x0,
                              minimizer_kwargs={'method': 'Nelder-Mead',
                                                'bounds': bnds,
                                                'args': (config_0,
                                                         dir_name,
                                                         paths)},
                              )
    elif opt_global == "shgo":
        result = shgo(_opti_hp_func,
                      bounds=bnds,
                      args=(config_0, dir_name, paths),
                      minimizer_kwargs={'method': 'Nelder-Mead', },
                      workers=workers,
                      options=opt_options,
                      )
    else:
        # Local optimization
        result = minimize(_opti_hp_func,
                          x0,
                          args=(config_0, dir_name, paths),
                          method='Nelder-Mead',
                          # tol=tolerance,
                          bounds=bnds,
                          # constraints=cons,
                          options={
                              # 'finite_diff_rel_step': .05,
                              "maxiter": max_iter,  # can take long!
                              "disp": True})
    return result, paths


def _opti_hp_func(x, conf, dir_name, paths, verbose=True, **kwargs):
    """
    Function for heat pump optimization, p_high and 4-component mixture

    Parameters
    ----------
    x : array/list, length 4
        p_high and the first three mole fractions.
    conf :  dict
        vales set here and deviating from dir_name.
    dir_name : str or dict
        either the path to the yaml file with all configuration parameters, or
        a dictionary with all values.
    paths: list of lists
        from function extract_optim_data_with_paths. Needed to reconstruct the
        config_0 dictionary from the optimizer x-list.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    conf_act = insert_optim_data(conf, x, paths)
    if "fractions" in conf_act["working_fluid"].keys():
        if conf_act["working_fluid"]["fractions"][-1] < 0:
            return -conf_act["working_fluid"]["fractions"][-1] * 10
    xn = np.array(x)
    if any(xn < 0):
        return - xn[xn < 0].sum() * 10

    try:
        hp_res = cb.hp_comp.heat_pump(dir_name,
                                      config=conf_act,
                                      verbose=verbose)

        p_low_actual = hp_res['output']["start"]["p_low"]
        p_low_min = hp_res['output']['config']['working_fluid']["p_low"]
        p_l_ratio = p_low_actual / p_low_min
        factor = 1
        if p_l_ratio <= 1:  # punish too low pressures
            factor = p_l_ratio * -10
        if verbose:
            for key, w_act  in hp_res['warnings'].items():
                if w_act.value  >0:
                    print(key,w_act, x)

    except Exception as e:
        print("Fehler aufgetreten:", type(e), e)
        return 5.9
    if any(ns.value > 0 for ns in hp_res['warnings'].values()):
        return sum(item.value for item in hp_res['warnings'].values())
    return -hp_res['COP'] * factor


# -------- ORC ------------------

def optimize_orc(dir_name, cop, q_dot_h, config_0, bounds_0, **kwargs):
    """
    ORC optimization, varying p_high and mixture composition etc.



    Parameters
    ----------
    dir_name : str or dict
        either the path to the yaml file with all configuration parameters, or
        a dictionary with all values.
    cop : float
        COP of the charging heat pump; is needed to secure steady state
        discharging.
    q_dot_h : float
        The heat flow from the high temperature storage to the working fluid
        in the evaporator.
    config_0 : dict
        vales set here and deviating from dir_name.
    bounds_0 : dictTYPE
        dictionary with the p_high bounds and the upper bounds of the first
        three molefractions. The remaining value is the difference to 1. It is
        checked that the sum is 1 and all vales are positive.

    **kwargs :
        - 'optimize_global': select optimization algorith "dif_evol", "bas_hop",
        or local minimizer (Nelson Mead).

    Returns
    -------
    result : OptimizeResult
        The result of the optimization.
    paths: list of lists
        from function extract_optim_data_with_paths. Needed to reconstruct the
        config_0 dictionary from the optimizer x-list.

    """

    x0, bnds, paths = extract_optim_data_with_paths(config_0, bounds_0)
    opt_global = kwargs.get('optimize_global', "dif_evol")
    max_iter = kwargs.get('maxiter', 200)
    workers = kwargs.get('workers', 1)
    opt_options =kwargs.get('opt_options',{})
    tolerance = 1e-4

    if opt_global == "dif_evol":

        result = differential_evolution(_opti_orc_func,
                                        bnds,
                                        args=(cop, q_dot_h, config_0,
                                              dir_name, paths),
                                        workers=workers,
                                        maxiter=max_iter,
                                        )
    elif opt_global == "bas_hop":
        result = basinhopping(_opti_orc_func,
                              x0,
                              minimizer_kwargs={'method': 'Nelder-Mead',
                                                'bounds': bnds,
                                                'args': (cop, q_dot_h, config_0,
                                                         dir_name,
                                                         paths)},
                              )
    elif opt_global == "shgo":
        result = shgo(_opti_orc_func,
                      bounds=bnds,
                      args=(cop, q_dot_h, config_0, dir_name, paths),
                      minimizer_kwargs={'method': 'Nelder-Mead',
                                        },
                      workers=workers,
                      options=opt_options,
                      )
    else:
        # Local optimization
        result = minimize(_opti_orc_func,
                          x0,
                          args=(cop, q_dot_h, config_0, dir_name, paths),
                          method='Nelder-Mead',
                          # tol=tolerance,
                          bounds=bnds,
                          # constraints=cons,
                          options={
                              # 'finite_diff_rel_step': .05,
                              "maxiter": max_iter,  # can take long!
                              "disp": True})
    return result, paths


def _opti_orc_func(x, cop, q_dot_h, conf, dir_name, paths, verbose=False, **kwargs):
    """
    Function for heat pump optimization, p_high and 4-component mixture

    Parameters
    ----------
    x : array/list, length 4
        p_high and the first three mole fractions.
    conf :  dict
        vales set here and deviating from dir_name.
    dir_name : str or dict
        either the path to the yaml file with all configuration parameters, or
        a dictionary with all values.
    paths: list of lists
        from function extract_optim_data_with_paths. Needed to reconstruct the
        config_0 dictionary from the optimizer x-list.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """

    conf_act = insert_optim_data(conf, x, paths)
    if "fractions" in conf_act["working_fluid"].keys():
        if conf_act["working_fluid"]["fractions"][-1] < 0:
            return -conf_act["working_fluid"]["fractions"][-1] * 10
    xn = np.array(x)
    if any(xn < 0):
        return - xn[xn < 0].sum() * 10

    # try:
    res_orc = cb.orc_comp.orc(dir_name, cop, q_dot_h,
                              config=conf_act)

    if verbose:
        print(res_orc["eta_th"], "\n", res_orc['warnings'], "\n", conf_act)

    # except Exception as e:
    #     print("Fehler aufgetreten:", type(e), e)
        return 5.09
    if res_orc["eta_th"] < 0:
        return - res_orc["eta"] * 10.00
    elif any(ns.value > 0 for ns in res_orc['warnings'].values()):
        return sum(item.value for item in res_orc['warnings'].values())

    return -res_orc["eta_th"]

# ---------------------------------------------------------------
# ------------------Carnot Battery/PTES -------------------------


def optimize_cb(dir_names, configs, bounds, **kwargs):
    """
    CB optimization, varying p_high and mixture composition etc.



    Parameters
    ----------
    dir_name : str or dict
        either the path to the yaml file with all configuration parameters, or
        a dictionary with all values.
    cop : float
        COP of the charging heat pump; is needed to secure steady state
        discharging.
    q_dot_h : float
        The heat flow from the high temperature storage to the working fluid
        in the evaporator.
    config_0 : dict
        vales set here and deviating from dir_name.
    bounds_0 : dictTYPE
        dictionary with the p_high bounds and the upper bounds of the first
        three molefractions. The remaining value is the difference to 1. It is
        checked that the sum is 1 and all vales are positive.

    **kwargs :
        - 'optimize_global': select optimization algorith "dif_evol", "bas_hop",
        or local minimizer (Nelson Mead).
        - 'workers': the number of workers for the differential evolution algorithm.
        Integer, optional, default value is 1.
        - 'maxiter': maximum iterations of optimizers, integer, default value is 200.

    Returns
    -------
    result : OptimizeResult
        The result of the optimization.
    paths: list of lists
        from function extract_optim_data_with_paths. Needed to reconstruct the
        config_0 dictionary from the optimizer x-list.

    """
    x = {}
    bnds = {}
    paths = {}

    x0 = []
    bnds_tot = []
    for key in ["hp", "orc"]:
        x[key], bnds[key], paths[key] = extract_optim_data_with_paths(
            configs[key], bounds[key])
        x0 += x[key]
        bnds_tot += bnds[key]

    opt_global = kwargs.get('optimize_global', "dif_evol")

    max_iter = kwargs.get('maxiter', 200)
    workers = kwargs.get('workers', 1)
    opt_options =kwargs.get('opt_options',{})
    args_act = (configs, dir_names, paths)

    if opt_global == "dif_evol":

        result = differential_evolution(_opti_cb_func,
                                        bnds_tot,
                                        args=args_act,
                                        workers=workers,
                                        maxiter=max_iter,
                                        x0=x0,
                                        )
    elif opt_global == "bas_hop":
        result = basinhopping(_opti_cb_func,
                              x0,
                              minimizer_kwargs={'method': 'Nelder-Mead',
                                                'bounds': bnds_tot,
                                                'args': args_act},
                              )
    elif opt_global == "shgo":
        result = shgo(_opti_cb_func,
                      bounds=bnds_tot,
                      args=args_act,
                      #minimizer_kwargs={'method': 'Nelder-Mead',},
                      workers=workers,
                      options=opt_options,
                      n = 999,
                      #sampling_method='sobol',
                      )
    else:
        # Local optimization
        result = minimize(_opti_cb_func,
                          x0,
                          args=args_act,
                          method='Nelder-Mead',
                          # tol=tolerance,
                          bounds=bnds_tot,
                          # constraints=cons,
                          options={
                              # 'finite_diff_rel_step': .05,
                              "maxiter": max_iter,  # can take long!
                              "disp": True})

    gc.collect()
    multiprocessing.active_children()  # leere Liste -> keine offenen Prozesse mehr?
    return result, paths


def _opti_cb_func(x, configs, dir_names, paths, verbose=False, **kwargs):
    """
    Function for heat pump optimization, p_high and 4-component mixture

    Parameters
    ----------
    x : array/list, length 4
        p_high and the first three mole fractions.
    conf :  dict
        vales set here and deviating from dir_name.
    dir_name : str or dict
        either the path to the yaml file with all configuration parameters, or
        a dictionary with all values.
    paths: list of lists
        from function extract_optim_data_with_paths. Needed to reconstruct the
        config_0 dictionary from the optimizer x-list.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    **kwargs : TYPE
        DESCRIPTION.

    Returns
    -------
    TYPE
        DESCRIPTION.

    """
    conf_act = {}

    for key in configs.keys():
        match key:
            case "hp":
                x_k = x[:len(paths["hp"])]
            case "orc":
                x_k = x[len(paths["hp"]):]
        conf_act[key] = insert_optim_data(configs[key], x_k, paths[key])
        if key == "hp" and conf_act[key]["working_fluid"]["fractions"][-1] < 0:
            return -conf_act[key]["working_fluid"]["fractions"][-1] * 10
    xn = np.array(x)
    if any(xn < 0):
        return - xn[xn < 0].sum() * 10

    try:
        rte, result = cb.cb_comp.cb_calc(dir_names,
                                         config=conf_act)

        if verbose:
            print(rte, "\n")

    except Exception as e:
        print("Fehler aufgetreten:", type(e), e)
        return 0.9
    if rte < 0:
        return -rte * 10.0

    return -rte

# --------------------------------------------------------


def extract_cb_conf_from_x(x, configs_ini, paths):
    conf_act = {}

    for key in configs_ini.keys():
        match key:
            case "hp":
                x_k = x[:len(paths["hp"])]
            case "orc":
                x_k = x[len(paths["hp"]):]

        conf_act[key] = insert_optim_data(configs_ini[key], x_k, paths[key])

    return conf_act




def extract_optim_data_with_paths(config, bounds):
    """
    Extracts all optimizable parameter values, along with their bounds and hierarchical paths,
    from a (possibly nested) configuration dictionary for optimization (e.g., heat pump application).

    Parameters
    ----------
    config : dict
        The (possibly nested) configuration dictionary containing initial values.
    bounds : dict
        Dictionary with bounds for all optimizable parameters, structured like config.

    Returns
    -------
    x0 : list
        A flat list of all initial parameter values to be optimized (e.g., p_high and mole fractions).
    bnds : list
        List of bounds tuples corresponding to all entries in x0.
    paths : list of lists
        Each entry is the hierarchical path (list of keys/indices) to the corresponding variable in config.
        Can be used to reconstruct the config dictionary from an x-list after optimization.

    Example
    -------
    >>> x0, bnds, paths = extract_optim_data_with_paths(config, bounds)
    """
    x0 = []
    bnds = []
    paths = []

    def _extract(conf_level, bounds_level, path):
        for key, val in conf_level.items():
            if key not in bounds_level:
                continue
            if isinstance(val, dict) and isinstance(bounds_level[key], dict):
                _extract(val, bounds_level[key], path + [key])
            elif key == "fractions":
                for ii, x_i in enumerate(val[:-1]):
                    x0.append(x_i)
                    bnds.append(tuple(bounds_level["fractions"][ii]))
                    paths.append(path + [key, ii])
            else:
                x0.append(val)
                bnds.append(tuple(bounds_level[key]))
                paths.append(path + [key])
    _extract(config, bounds, [])
    return x0, bnds, paths

# ---------------------------------------------------------------


def insert_optim_data(config, x, paths):
    """
    Reconstructs a configuration dictionary from a flat list of optimized values
    and corresponding paths, including special handling for mole fractions.
    For back conversion after optimization (scipy.optimize).

    Parameters
    ----------
    config : dict
        The original configuration dictionary (will be deep-copied).
    x : list or array
        The optimizer parameter vector (e.g., for p_high and first n-1 mole fractions).
    paths : list of lists
        The list of hierarchical paths as produced by extract_optim_data_with_paths;
        describes where in config each value from x should be inserted.

    Returns
    -------
    conf_new : dict
        A new dictionary with the updated parameter values from x inserted.
        For fields called 'fractions', the last value is set to 1 minus the sum
        of the others (to ensure the fractions sum to 1).

    Example
    -------
    >>> conf_new = insert_optim_data(config, x, paths)
    """

    conf_new = deepcopy(config)
    # Key: Pfad bis 'fractions', Value: Liste von (Pfad, x-Index)
    fractions_groups = {}
    for i, path in enumerate(paths):
        if len(path) >= 2 and path[-2] == "fractions":
            # Neu: bis VOR 'fractions' (also z.B. ["working_fluid"])
            grp = tuple(path[:-2])
            if grp not in fractions_groups:
                fractions_groups[grp] = []
            fractions_groups[grp].append((path, i))

    # Normale Parameter
    for val, path in zip(x, paths):
        if len(path) >= 2 and path[-2] == "fractions":
            continue  # fractions werden gesammelt behandelt
        conf_level = conf_new
        for p in path[:-1]:
            conf_level = conf_level[p]
        conf_level[path[-1]] = val

    # fractions-Gruppen: Korrigierter Zugriff
    for grp, path_indices in fractions_groups.items():
        # n-1 optimierbare mole fractions stehen im Pfad
        group_length = len(path_indices) + 1
        fraction_values = [x[idx] for _, idx in sorted(
            path_indices, key=lambda xx: xx[0][-1])]
        last_val = 1.0 - sum(fraction_values)
        full_fractions = fraction_values + [last_val]
        # Zugriff bis zum richtigen dict
        conf_level = conf_new
        for p in grp:
            conf_level = conf_level[p]
        conf_level["fractions"] = full_fractions

    return conf_new


def parse_species_string(species_str):
    """
    Converts a species string in format 'Propane * Butane * Pentane * Hexane'
    into a list without spaces and asterisks.

    Parameters
    ----------
    species_str : str
        Species string with asterisk separation

    Returns
    -------
    list
        List of species names
    """
    if isinstance(species_str, str):
        return [species.strip() for species in species_str.split('*')]
    return species_str  # If already a list


def extract_cycle_column_names_from_config(config_dict, paths):
    """
    Extracts column names for optimization variables based on paths.
    Returns a single flat list with prefixed names for non-fraction variables.

    Parameters
    ----------
    config_dict : dict
        Dictionary with configuration data (contains species names for fractions)
        Structure: config_dict["working_fluid"]["species"] = "Propane * Butane * ..."
    paths : list
        Dictionary with paths for each optimization 

    Returns
    -------
    list
        Single flat list of column names for all optimization variables
    """
    all_names = []

    
    for path in paths:
        if len(path) >= 2 and path[-2] == "fractions":
            # For fractions: Extract species names from config_dict (no prefix)
            # Find the working_fluid section that contains the species
            species_list = None

            # Look for species in the path hierarchy (usually under working_fluid)
            for key in path[:-2]:  # Go through path up to 'fractions'
                if key == "working_fluid" and "working_fluid" in config_dict:
                    if "species" in config_dict["working_fluid"]:
                        species_list = parse_species_string(
                            config_dict["working_fluid"]["species"])
                        break

            # If species found, use species name, otherwise use generic name
            if species_list:
                fraction_idx = path[-1]
                if fraction_idx < len(species_list):
                    all_names.append(
                        f"{species_list[fraction_idx]}_fraction")
                else:
                    all_names.append(f"fraction_{fraction_idx}")
            else:
                all_names.append(f"fraction_{path[-1]}")
        else:
            # For normal parameters: Use group prefix + parameter name
            all_names.append(f"{path[-1]}")

    return all_names



def extract_column_names_from_config(config_dict, paths_dict):
    """
    Extracts column names for optimization variables based on paths.
    Returns a single flat list with prefixed names for non-fraction variables.

    Parameters
    ----------
    config_dict : dict
        Dictionary with configuration data (contains species names for fractions)
        Structure: config_dict["working_fluid"]["species"] = "Propane * Butane * ..."
    paths_dict : dict
        Dictionary with paths for each optimization group (e.g. "hp", "orc")

    Returns
    -------
    list
        Single flat list of column names for all optimization variables
    """
    all_names = []

    for group_key, paths in paths_dict.items():
        for path in paths:
            if len(path) >= 2 and path[-2] == "fractions":
                # For fractions: Extract species names from config_dict (no prefix)
                # Find the working_fluid section that contains the species
                species_list = None

                # Look for species in the path hierarchy (usually under working_fluid)
                for key in path[:-2]:  # Go through path up to 'fractions'
                    if key == "working_fluid" and "working_fluid" in config_dict:
                        if "species" in config_dict["working_fluid"]:
                            species_list = parse_species_string(
                                config_dict["working_fluid"]["species"])
                            break

                # If species found, use species name, otherwise use generic name
                if species_list:
                    fraction_idx = path[-1]
                    if fraction_idx < len(species_list):
                        all_names.append(
                            f"{species_list[fraction_idx]}_fraction")
                    else:
                        all_names.append(f"fraction_{fraction_idx}")
                else:
                    all_names.append(f"fraction_{path[-1]}")
            else:
                # For normal parameters: Use group prefix + parameter name
                all_names.append(f"{group_key}_{path[-1]}")

    return all_names


"""
Example for usage of the Turbo Design Tool
"""
import carbatpy as cb
import sys
import time
import datetime

from carbatpy.optional.spp_machines import Compressor, Turbine, require_spp_machines

require_spp_machines()

#%%
# Compressor
# Initiate Compressor
print(datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))

#%%
Results = Compressor()
# Example from Repository
# Create inputs for the Tool
InletTemperature = 300           # Inlet temperature in K
InletPressure = 1e5              # Inlet pressure in Pa
OutletPressure = 30e5             # Desired outlet pressure in Pa
WorkingFluid = "ethane * propane * pentane * hexane"  # Working fluid as named in REFPROP
WorkingFluidComposition = [0.15, 0.75, 0.05, 0.05]    # Working fluid composition in mole fractions
InletVolumeFlow = 1.5              # Inlet volume flow in m^3/s

# Calling the design method
OutletState, MachineData = Results.designCompressor(WorkingFluid, 
                                                    WorkingFluidComposition, 
                                                    InletPressure, 
                                                    InletTemperature, 
                                                    OutletPressure, 
                                                    InletVolumeFlow, 
                                                    GearedCompressor=False,
                                                    Plot=True,
                                                    EfficiencyConvergence=5e-2
                                                    )

# Results:
    # OutletState: working fluid state at the compressor outlet as a carbatpy state object
    # MachineData: dictionary containing the details of the compressor design, used by further methods
#%%
# Outletstate
print(OutletState.state)

# Total efficiency in MachineData at the last stage:
print(Results.IsentropicEfficiency[-1])
print(Results.PolytropicEfficiency)

#%%
# Turbine
# Create inputs for the Tool
InletTemperature = 300+273.15
InletPressure = 20e5
OutletPressure = 1e5 
WorkingFluid = "ethane * propane * pentane * hexane"
WorkingFluidComposition = [0.15, 0.75, 0.05, 0.05]
MassFlow = 1
TS = cb.CB_DEFAULTS['Fluid_Defaults']['TRANS_STRING']

myfluid = cb.init_fluid(WorkingFluid, WorkingFluidComposition)
InletState = myfluid.set_state([InletTemperature, InletPressure], "TP", 
                               wanted=TS,output="FluidState")

# Class initialisation
Results_turbine = Turbine()

# Calling the design method
OutletState_turbine, MachineData_turbine = Results_turbine.designTurbine(WorkingFluid, 
                                                 WorkingFluidComposition, 
                                                 InletState, 
                                                 OutletPressure, 
                                                 MassFlow, 
                                                 Plot=True)

#%%
# Outletstate
print(OutletState_turbine.state)

# Total efficiency in MachineData at the last stage:
print(Results_turbine.IsentropicEfficiency[-1])
print(Results_turbine.PolytropicEfficiency)



# -*- coding: utf-8 -*-
"""
Created on Thu Jan  8 18:04:59 2026

@author: atakan
Universität Duisburg-Essen, Germany

In the framework of the Priority Programme: "Carnot Batteries: Inverse Design from
Markets to Molecules" (SPP 2403)
https://www.uni-due.de/spp2403/
https://git.uni-due.de/spp-2403/residuals_weather_storage

"""
# carbatpy/optional/spp_machines.py
from __future__ import annotations

from importlib.util import find_spec

HAS_SPP_MACHINES = find_spec("spp_machines") is not None

if HAS_SPP_MACHINES:
    # Compressor
    from spp_machines.inverseturbomachinerydesigner.RadialCompressorDesignTool import (
        inverseCompressorDesignTool as Compressor,
    )

    # Turbine
    from spp_machines.radialturbinedesigntool.InverseTurbineDesignTool import (
        InverseTurbineDesignTool as Turbine,
    )
else:
    Compressor = None
    Turbine = None


def require_spp_machines() -> None:
    """Raise a helpful error if spp-machines is not installed."""
    if not HAS_SPP_MACHINES:
        raise ImportError(
            "Optional dependency 'spp-machines' is not installed. "
            "Install it in the same environment as carbatpy, e.g. "
            "'pip install spp-machines' or (Poetry) 'poetry install --with spp_machines'."
        )


# -*- coding: utf-8 -*-
"""
Created on Thu Aug  1 12:35:02 2024

@author: atakan
"""
import json
import yaml
import carbatpy as cb

input_hp = {
    'process': {
        "compressor": "Flow_device",
        "condenser": "StaticHeatExchanger",
        "throttle": "Flow_device",
        "evaporator": "StaticHeatExchanger",

    },
    "species_all": {'working_fluid': "cycle",
               'cold_storage': "source-sink",
               'hot_storage': "source-sink",
               },
    'compressor': {
        "calc_type": "const_eta",
        "eta_s": 0.7,
        "species": {"working_fluid":{"in": "evaporator",
                                   "out":"condenser"},},
        "dt_superheat": 5,
        
    },
    'throttle': {
        "calc_type": "const_h",
        "species": {"working_fluid":{"in": "condenser" ,
                                   "out":"evaporator"},},
    },
    'condenser': {
        "calc_type": "const",
        "species": {"working_fluid":{"in":"compressor",
                                   "out":"throttle"}, 
                  "hot_storage":{"in":"temp_low",
                                 "out":"temp_high"}},
        "dt_min": 3.,

    },
    'evaporator': {
        "calc_type": "const",
        "species": {"working_fluid":{"in":"throttle",
                                   "out":"compressor"}, 
                  "cold_storage":{"in":"temp_high",
                                 "out":"temp_low"}},
        "dt_min": 3.,
    },
    'cold_storage': {
        "species": "Methanol",
        "fractions": [1.0],
        "p_low": 5e5,
        "temp_low": 250.0,
        "temp_high": "ambient"
    },
    'hot_storage': {
        "species": "Water",
        "fractions": [1.0],
        "p_low": 2e5,
        "temp_low": "ambient",
        "temp_high": 360.0
    },
    'working_fluid': {
        "species": "Propane * Butane * Pentane * Hexane",
        "fractions": [
            7.71733787e-01,
            2.22759488e-02,
            1.78685867e-01,
            0.027304397199999997],
        
        "temp_low": "ambient",
        "temp_high": 360.0,
        "p_low": 1.28250708e+05,
        "p_high": 1.37548728e+06,

        "optimize": "None",
        "setting": "initial",
        
    },
}


input_hex = {
    
    'condenser': {
        
        'model': "StaticHeatExchanger",
        "calc_type": "const",
        "species": {"working_fluid":{"in":"temp_high",
                                   "out":"temp_low"}, 
                  "hot_storage":{"in":"temp_low",
                                 "out":"temp_high"}},
        "dt_min": 3.,
        "q_dot": 1000.,
        "overall_u":100,
        'area': 1,

    },
    'hot_storage': {
        "species": "Water",
        "fractions": [1.0],
        "p_low": 2e5,
        "temp_low": "ambient",
        "temp_high": 360.0,
        'props':"REFPROP",
        
    },
    'working_fluid': {
        "species": "Propane * Butane * Pentane * Hexane",
        "fractions": [
            7.71733787e-01,
            2.22759488e-02,
            1.78685867e-01,
            0.027304397199999997],
        
        "temp_low": "ambient",
        "temp_high": 360.0,
        "p_low": 1.28250708e+05,
        "p_high": 1.37548728e+06,

        "optimize": "None",
        "setting": "initial",
        'props':"REFPROP",
        
    },
}

class Val():
    """
    Class to store and read the *input* dictionary values and variables for a heat pump.

    Best is to set them in a yaml or json file and read them with the appropriate
    function. The default place to search for hp-input-dictvariables is in the
    data directory.

    Part of carbatpy.
    """

    def __init__(self, variables_dict=None):
        if variables_dict:
            for key, value in variables_dict.items():
                setattr(self, key, value)

    def to_file(self, fname=None):
        if fname:
            with open(fname, "w") as file:
                if fname.find("json") > 0:
                    json.dump(self.__dict__, file, indent=4)
                elif fname.find("yaml") > 0:
                    yaml.dump(self.__dict__, file)
                else:
                    raise NotImplementedError(f"File type not implemeted {fname}")
        else:
            print("No file name given")

    @classmethod
    def from_file(cls, fname=None):
        instance = cls.__new__(cls)
        if fname:
            with open(fname, "r") as file:
                if fname.find("json") > 0:
                    data = json.load(file)
                elif fname.find("yaml") > 0:
                    data = yaml.safe_load(file)
                else:
                    raise NotImplementedError(f"File type not implemeted {fname}")
            for key, value in data.items():
                setattr(instance, key, value)
        else:
            print("No file name given")
        return instance


def find_key(a_dict, fl_name, t_amb =cb.CB_DEFAULTS['General']["T_SUR"]):
    for key in a_dict["process"].keys():
        di2 = a_dict[key]
        if fl_name in di2["species"].keys():
            in_out = di2["species"][fl_name]
            t_both= {}
            for key, value in in_out.items():
                t_act=  a_dict[fl_name][value]
                if t_act =="ambient":
                    t_act = t_amb
                t_both[key]=t_act
            #T_in, T_out = a_dict[fl_name][in_out["in"]], a_dict[fl_name][in_out["out"]]
            return t_both


if __name__ == "__main__":
    
    
    RES_DIR = cb.CB_DEFAULTS["General"]["RES_DIR"]

    fn0 = RES_DIR + "\\test_input3.yaml"

    #     mein = Val(input_hp)

    actuell = Val(input_hp)
    actuell.to_file(fn0)
    neu = Val.from_file(fn0)
    print (f'{find_key(input_hp, "cold_storage")}, hot: {find_key(input_hp, "hot_storage")}')
    with open(fn0, "r") as file:
        if fn0.find("json") > 0:
            in_dict = json.load(file)
        elif fn0.find("yaml") > 0:
            in_dict = yaml.safe_load(file)



# -*- coding: utf-8 -*-
"""
Created on Tue Jul 30 16:15:25 2024
structure to create Surrogate models
01.08.2024: for multilayer perceptrons (have proven themselves)
needs a labeled DataFrame with training data

@author: welp
"""
import carbatpy as cb
import pandas as pd
import joblib
import os
import pickle
import numpy as np
import yaml

from sklearn.preprocessing import MinMaxScaler
from sklearn.model_selection import train_test_split
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import RandomizedSearchCV

from datetime import datetime

class Surrogate:
    
    def __init__(self, title):
        self.title = title
                
        
    def _make_xy(self, DF):
        '''
        creates feature and target array and dataframe

        '''
        
        #create X and y from feature and target list
        self.DF_x = DF[self.features_list]  
        self.DF_y = DF[self.targets_list]
     
        self.x_data_arr = self.DF_x.to_numpy(copy=True)
        self.y_data_arr = self.DF_y.to_numpy(copy=True)
        
    def train_surrogate(
            self, DF, features_list, targets_list, split = 0.2, 
            random_state = 42, hypo="def_hyperparameter.yaml", verbose=False):
        """
        train MLP surrogate from dataframe with chosen features and targets,
        uses minmaxscaler and 'r2', 'neg_root_mean_squared_error' for scoring,
        the latter for refitting
        
        Parameters
        ----------
        DF : TYPE
            DESCRIPTION.
        features_list : TYPE
            DESCRIPTION.
        targets_list : TYPE
            DESCRIPTION.
        split : TYPE, optional
            DESCRIPTION. The default is 0.2.
        random_state : TYPE, optional
            DESCRIPTION. The default is 42.
        hypo : TYPE, optional
            DESCRIPTION. The default is "def_hyperparameter.yaml".

        Returns
        -------
        Y_test : TYPE DataFrame
            DESCRIPTION. test targets
        Y_pred : TYPE
            DESCRIPTION. predicted targets

        """
        
        # create train and test set
        self.features_list = features_list
        self.targets_list = targets_list
        self.split = split
        self.random_state = random_state
        
        self.scaler_x = MinMaxScaler()
        self.scaler_y = MinMaxScaler()
        self._make_xy(DF)
    
        X_train, X_test, y_train, y_test = train_test_split(
            self.x_data_arr, self.y_data_arr, test_size=self.split, random_state=self.random_state)
        self.x_range = [np.min(X_train, axis=0), np.max(X_train, axis=0)]
        X_train = self.scaler_x.fit_transform(X_train)
        X_test = self.scaler_x.transform(X_test)
    
        y_train= self.scaler_y.fit_transform(y_train)
        y_test = self.scaler_y.transform(y_test)
    
        with open(hypo, 'r') as file:
            self.parameters = yaml.safe_load(file)
            
        # Define and start hyperparameter optimization
        if hypo ==  "def_hyperparameter.yaml":
            full_set = []
            for i,nhl in enumerate(self.parameters["hidden_layers"]):
                for j, n in enumerate(self.parameters["neurons"]):
                    NN = [n] * nhl
                    NN = list(NN)
                    full_set.append(NN)
            del self.parameters['hidden_layers']
            del self.parameters['neurons']
            self.parameters['hidden_layer_sizes'] = full_set
                
        # Create MLP for all targets
        model = MLPRegressor(max_iter=2000, n_iter_no_change=30)
    
        RS = RandomizedSearchCV(model, self.parameters, 
                            scoring=('r2', 'neg_root_mean_squared_error'), 
                            refit='neg_root_mean_squared_error', n_iter=100,
                            n_jobs = -1, cv=3)
        
        RS.fit(X_train, y_train)
        
        self.RS = RS
        self.model = RS.best_estimator_
        
        if verbose == True:
            print('best parameters: ', RS.best_params_)
            print('best estimator: ', RS.best_estimator_)
            print('best score: ', RS.best_score_)
        
        y_pred = RS.predict(X_test)

        y_pred = self.scaler_y.inverse_transform(y_pred)
        y_test = self.scaler_y.inverse_transform(y_test)
        Y_pred = pd.DataFrame(y_pred, columns=targets_list)
        Y_test = pd.DataFrame(y_test, columns=targets_list)
        return Y_test, Y_pred

        
    def save(self, path="default"):
        """
        save model

        Parameters
        ----------
        path : TYPE, optional
            DESCRIPTION. The default is CARBATPY_RES_DIR

        Returns
        -------
        None.

        """
        
        elements = '-'.join(self.features_list)
        # Get the current date and time
        current_time = datetime.now()
        # Format the date and time as a string
        timestamp = current_time.strftime("%Y-%m-%d_%H-%M-%S")
        if path == "default":
            file_out = cb.CB_DEFAULTS["General"]["RES_DIR"] +'\\surrogates\\MLP-R\\' + elements + '\\' + timestamp + '\\'
        else:
            file_out = path + elements + '\\' + timestamp + '\\'
        try:
            os.makedirs(file_out)
        except: 
            pass    
        
        # Save results
        joblib.dump(self.RS, file_out + 'MLP_RS.pkl')
        joblib.dump(self.RS, file_out + 'MLP_RS.sav')
        joblib.dump(self.scaler_x, file_out + 'MLP_scaler_x.pkl')
        joblib.dump(self.scaler_y, file_out + 'MLP_scaler_y.pkl')
        joblib.dump(self.model, file_out + 'mlp.pkl')  


        self.DF_x.to_csv(file_out + 'X.txt', index=False)
        self.DF_y.to_csv(file_out + 'y.txt', index=False)

        # Creates output file with information and results
        filename = "MLP_Randomized_search_parameter.txt"
        file = open(file_out +  filename, "w")
        for key in range(len(list(self.parameters.keys()))):
            text = list(self.parameters.keys())[key] + ': ' + str(self.parameters[list(self.parameters.keys())[key]])+'\n'
            file.write(text)
        file.close()

        with open(file_out + 'features_list.pkl', 'wb') as file:
            pickle.dump(self.features_list, file)
            
        with open(file_out + 'targets_list.pkl', 'wb') as file:
            pickle.dump(self.targets_list, file)    
        
        with open(file_out + 'random_state.pkl', 'wb') as file:
            pickle.dump(self.random_state, file)
        
        with open(file_out + 'x_range.pkl', 'wb') as file:
            pickle.dump(self.x_range, file)

        # Ergebnisse in eine Textdatei schreiben
        filename = 'MLP_randomized_search_results.txt'
        with open(file_out + filename, "w") as file:
            file.write("Best parameter: {}\n".format(self.RS.best_params_))
            file.write("Best score: {}\n".format(self.RS.best_score_))
            
    def load(self, path):
        # loading ML models
        self.model = joblib.load(path + 'mlp.pkl')
        self.scaler_x = joblib.load(path + 'MLP_scaler_x.pkl')
        self.scaler_y = joblib.load(path + 'MLP_scaler_y.pkl')

        # Loading features and targets
        with open(path + 'features_list.pkl', 'rb') as file:
            self.features_list = pickle.load(file)
            
        with open(path + 'targets_list.pkl', 'rb') as file:
            self.targets_list = pickle.load(file) 
        
        with open(path + 'random_state.pkl', 'rb') as file:
            self.random_state = pickle.load(file)
            
        with open(path + 'x_range.pkl', 'rb') as file:
            self.x_range = pickle.load(file)
    
    def predict(self, DF_new_x):
        """
        use existing surrogate to predict new data

        Parameters
        ----------
        DF_new_x : TYPE
            DESCRIPTION.

        Raises
        ------
        ValueError
            DESCRIPTION. needs to contain features of model
        
            DESCRIPTION. does not extrapolate

        Returns
        -------
        y : TYPE
            DESCRIPTION. target results
        DF_y : TYPE
            DESCRIPTION. target results

        """
        
        try:
            DF_x = DF_new_x[self.features_list] 
        except:
            raise ValueError("features in DF_new_x not in features_list of model")

        X = DF_x.to_numpy(copy=True)
        if np.all(np.all((X >= self.x_range[0]) & (X <= self.x_range[1]), axis=0)) != True:
            raise ValueError(f"""x_predict is outside of data range used for training
                             range of training data:
                                  {self.features_list}
                             min: {self.x_range[0]}
                             max: {self.x_range[1]}""")
        try:
            X_scaled = self.scaler_x.transform(X)
            y_scaled = self.model.predict(X_scaled)
            y = self.scaler_y.inverse_transform(y_scaled)
        except:
            raise 
        DF_y = pd.DataFrame(y)
        DF_y.columns = self.targets_list
        return y, DF_y
        
    
if __name__ == "__main__":
    mode = 1  # 1: train new, 2: load existing
    
    if mode == 1:
        # first test case
        compressor = Surrogate("piston compressor")   
        DF = pd.read_csv(cb.CB_DEFAULTS['General']['CB_DATA'] + '\\Example_compressor_data.csv')    
        DF = DF[DF.counter < 100]       # filter failed results, iteration stops after 100 trials
        DF = DF[DF.is_eff != 1]         # eliminates 1 (invalid fluid properties)
    
        # choose features and targets
        features_list =  ['p_ve', 'T_e', 'v_e']
        targets_list = ['is_eff', 'degree_delivery', 'T_aus']
        
        compressor.train_surrogate(DF, features_list, targets_list)
        
        #compressor.save()  # not activated (will trash the discspace)
        
        
    elif mode == 2:
        my_new_compressor = Surrogate("my new compressor")
        my_new_compressor.load(cb.CB_DEFAULTS['General']['CB_DATA'] + '\\def_hyperparameter.yaml')
        x_predict = np.array([[7, 290, 0.2], [3, 305, 0.3]])
        x_predict = pd.DataFrame(x_predict)
        x_predict.columns = ['p_ve', 'T_e', 'v_e']
        y, y_DF = my_new_compressor.predict(x_predict)
    
    



# -*- coding: utf-8 -*-
"""
Local optimization of a heat pump.

pressure levels and working fluid mixture composition is varied.

Part of carbatpy

Created on Tue Jul 23 13:21:18 2024

@author: atakan
"""

import copy
import json
from time import time
# import src.models.fluids.fluid_props as fprop
import numpy as np
from scipy.optimize import minimize, minimize_scalar, Bounds  # root, root_scalar
import matplotlib.pyplot as plt
import carbatpy as cb


def p_var_for_opt_hp(pressures_x,
                     fixed_p=None,
                     verbose=False,
                     **kwargs):
    """Calculates -cop for given pressure levels and mole fractions -- for minimization.

    For set_fl_state the value fixed_p['p_select']["setting"] is important. If it is
    'auto' nothing is changed e.g. {"optimize": "no", "setting": "auto", 
     "p_low": pressures_x[0],"p_high": pressures_x[1]} and an evaluation is calculated for
    the given parameters. Anything else uses the values in pressures_x!


    Parameters
    ----------
    pressures_x : np.array
        the values to be varied here p_low, p_high and all but the last mole fraction.
    fixed_p : dictionary, optional
        all other parameters needed for the heat pump calculatiob (see there). The default is None.
    verbose : TYPE, optional
        DESCRIPTION. The default is False.
    **kwargs : TYPE
        if evalute = True, nit the cop is returned, but the heat pump instance.

    Returns
    -------
    TYPE
        -cop or HeatPump instance at the desired (optimal) point.

    """

    evaluate = kwargs.get("evaluate", None)
    fixed = copy.copy(fixed_p)
    n_sp = len(pressures_x) - 1
    mole_fract = np.zeros(n_sp)
    mole_fract[:-1] = pressures_x[2:]
    if not 0 < mole_fract[:-1].sum() < 1:
        return 160 * abs( mole_fract[:-1].sum())
    x_last = 1 - mole_fract.sum()
    mole_fract[-1] = x_last
    fixed["fluids_all"]["WORKING"][1] = mole_fract

    
    if evaluate:
        p_select = {"optimize": "no",  # is unused!
                    "setting": "auto",  # zjis value is important in set_fl_state
                    "p_low": pressures_x[0],
                    "p_high": pressures_x[1]}
    else:

        p_select = {"optimize": "p and x",  # is unused!
                    "setting": "fixed-p",
                    "p_low": pressures_x[0],
                    "p_high": pressures_x[1]}
    fixed["P_WORKING"] = p_select

    hp_act = cb.hp_simple.HeatPump(fixed)
    cop = hp_act.calc_heat_pump(verbose=False)
    if verbose:
        formatted_list = [f"{num:.2f}" for num in [*pressures_x]]
        print("\nIn opt:", ", ".join(formatted_list))
        print(hp_act.warning, cop, p_select, fixed["P_WORKING"], fixed["fluids_all"]["WORKING"])
    if evaluate is not None:
        hp_act.hp_plot()
        hp_act.save_to_file()
        return copy.copy(hp_act)
    if len(hp_act.warning) == 0:  # for minimization
        return -cop
    return 100 * cop


def my_save(a, fn=None):
    if fn is None:
        fn = cb.CB_DEFAULTS["General"]["RES_DIR"]+"\\opti-test.yml"
    a = cb.hp_simple.HpVal(a)
    a.save_to_file(fn)

def set_res_to_fixed(hp_act, values, names, n_opti_else=2):
    """
    Change the fxed points to the optimal values
    
    Only for optimization of the two working fluid pressures (must be first),
    another value (actually dT_Superheat) and the mole fractions.

    Parameters
    ----------
    hp_act : HeatPump
        the actual heat pump instance.
    values : list/array
        the five values to be set.
    names : list of strings
        the names of the values to be set, as given in the fixed_points
        dictionary.
    n_opti_else : integer, Optional
        number of other objectives (besiges species mole fractions) within
        optimization. Default value is 3.

    Returns
    -------
    dictionary
        with fixed points.

    """
    
    n_species =len(values) - n_opti_else + 1
    
    mole_fract =np.zeros(n_species)
    mole_fract[:-1] = values[n_opti_else:]
    mole_fract[-1] = 1 - mole_fract.sum()
    hp_act.fixed_points["fluids_all"]["WORKING"][1] = mole_fract
    
    for ii, v_act in enumerate(values[:-n_species+1]):
        if ii <2:
            hp_act.fixed_points['P_WORKING'][names[ii]]=v_act
        else:
            hp_act.fixed_points[names[ii]]=v_act
    return copy.copy(hp_act.fixed_points)
    

if __name__ == "__main__":
    # -----------------------------------------------------------------------------------
    # Set the starting point from the standard heat pump file and calculate pressures etc.
    
    fn0= cb.CB_DEFAULTS["General"]["CB_DATA"]+ "\\hp-input-dict_opt_px.json"
    with open(fn0,"r") as file:
              dict_ini = json.load(file)
    inputs = cb.hp_simple.HpVal.load_from_file(fn0)
    fixed_points = inputs.to_dict()
    act_points = copy.copy(fixed_points)
    hpn = cb.hp_simple.HeatPump(fixed_points)
    print(fixed_points, hpn.evaluation)
    
    cop_n = hpn.calc_heat_pump(verbose=True)
    print(f"Initial cop: {cop_n:.3f}")
    
    ps = ["p_low", "p_high"]
    # this should be used when pressures are set and not calculated
    p_w = [hpn.fixed_points["P_WORKING"][key] for key in ps]
    # here we use the previously calculated values
    p_w = [hpn.fixed_points[key] for key in ps]
    fl_all = hpn.fixed_points["fluids_all"]
    pressures_act = [*p_w,
                     *fl_all["WORKING"][1][:-1]]

    

   # set boundaries and tolerance
    b_min = [.8e5, 5e5, 0, 0, 0]
    b_max = [4e5, 19e5, 1, 1, 1]
    bound_act = Bounds(lb=b_min,
                       ub=b_max)
    tolerance = 5e-3 # should be good enough
    max_iter = 1222
    t0 = time()

    result = minimize(p_var_for_opt_hp,
                      pressures_act,
                      args=(
                            fixed_points,
                            {"verbose": True}),
                      method='Nelder-Mead',
                      tol=tolerance,
                      bounds=bound_act,
                      options={"maxiter": max_iter, # can take long!
                               "disp": True})
    t1 = time()
    directory = cb.CB_DEFAULTS["General"]["RES_DIR"]
    filename = directory+"\\hp_opti"
    print(result, "\ntime:", t1-t0)

    # The next lines are new. Here the optimal point is re-evaluated and plotted
    opt_names =["p_low", "p_high"]
    act_points = set_res_to_fixed(hpn, result.x, opt_names)
    hp_opt = cb.hp_simple.HeatPump(act_points)
    cop = hp_opt.calc_heat_pump(verbose=False)
    hp_opt.hp_plot()
    
    
    hp_opt.opti_result=result
    
    hp_opt.save_to_file(filename+"2p.json")
    hp_opt.save_to_file(filename+"2p.yaml")
    hp_opt_d = hp_opt.to_dict()
    print(hp_opt_d)
    print(fixed_points==act_points)
    dict_final ={}
    for key in dict_ini:
        dict_final[key]=hp_opt_d["fixed_points"][key]
    final = cb.hp_simple.HpVal(dict_ini)
    final.save_to_file(filename+"-opti-input2p.yaml")
    
    print("Warning=", hp_opt.warning)
    print("COP=" ,hp_opt.evaluation['COP'])
    print("dT min Condenser=" ,hp_opt.evaluation['condenser']['dT-mean'])
    print("dT min evaporator=" ,hp_opt.evaluation['evaporator']['dT-mean'])
    print("P min=" ,hp_opt.evaluation['p_low'])
    print("P max=" ,hp_opt.evaluation['p_high'])
    print("xCO2=",hp_opt.fluids_all['WORKING'][1][0])
    print("x Butane=",hp_opt.fluids_all['WORKING'][1][1])
    print(result.x)
    



#!/usr/bin/env python
# coding: utf-8

# ## Fluid Properties
# Using carbatpy, the default is evaluating fluid properties with REFPROP, but along initializing a fluid you can also set props="TREND". You should have installed the package you want to use, they are not part of carbatpy!

# In[9]:


import carbatpy as cb


# Select a fluid, calculate a state using the default REFPROP and print it out:

# In[10]:


FLUID = "Propane * Pentane"
    
comp = [.5, .5]
myFluid = cb.init_fluid(FLUID, comp)

quality = 0.5       # is interpreted by carbatpy as mol/mol to calculate state
# but given back in string as kg/kg
st0 = myFluid.set_state([300., quality], "TQ")
myFluid.print_state()


# ## Properties and Units
# The properties can be accessed from a dictionary. **Units are SI** (mass base: K, kg. Pa, m), only composition is always in **mole fractions**.

# In[11]:


myFluid.val_dict


# or using the properties (FluidState Class):

# In[12]:


myFluid.properties.enthalpy


# You are interested in transport properties? First check the Default settings:

# In[13]:


cb.CB_DEFAULTS


# Then we need the TRANS_STRING the default is the THERMOSTRING:

# In[14]:


TS = cb.CB_DEFAULTS['Fluid_Defaults']['TRANS_STRING']
state = myFluid.set_state([300, 1e5], "TP", TS)


# In[15]:


myFluid.val_dict


# In[ ]:





# In[ ]:





# In[ ]:





# ## Return Value
# By default the returned output (state) is a list (in the same order, as above, but without x,y, z), which may be more convenient to use, but the order of the properties after the internal energy may change in later versions.
# by setting the keyword "output" = "FluidState", you can get an instanse of FluidState as return value and you can access the properties by name (here: spec. enthalpy):
# 
# 

# In[19]:


my_prop = myFluid.set_state([300, 1e5], "TP",output="FluidState")
print("FluidState:\n",my_prop.state_to_dict(),f"\nEnthalpy: {my_prop.enthalpy}\n")
print(f"\nstate:\n {state}")


# ## TREND
# If you installed it, you can use it. It needs some additional input (see TREND manual). Although we initialize with "TP" (temperature, pressure), you can use it with other combinations also, if they are installed in TREND. The units are as said above.
# Transport property evaluation will be implemented soon, but the calculated values differ from those in REFPROP and currently mixtur transport properties cannot be evaluated.
# Both cannot handle two-phase fluidtransport properties!

# In[ ]:


TREND = cb.CB_DEFAULTS['Fluid_Defaults']['TREND']

if TREND["TREND_INSTALLED"]:
    trend_dll = TREND["TREND_DLL"]
    trend_path = TREND["TREND_PATH"]
    
tr_dict = {"Input": "TP",
               'calctype': "H",
               'fluids': FLUID,
               "moles": comp,
               "eos_ind": [1, 1],
               'mix_ind': 1,
               'path': trend_path,
               'unit': 'specific',
               'dll_path': trend_dll}
my_trend_fluid = cb.init_fluid(FLUID, comp, props="TREND", args= tr_dict)


# In[ ]:


state = my_trend_fluid.set_state([300, 1e5], "TP")
my_trend_fluid.val_dict


# In[ ]:







# -*- coding: utf-8 -*-
"""

Read and load an instance of a class from/to a file

from chatgpt 

Created on Thu Jul 25 13:40:23 2024

@author: atakan
"""

import json
import yaml

def save_instance_to_file(instance, file_path):
    # Wandle die Instanz in ein Dictionary um
    data = instance.__dict__.copy()

    if file_path.endswith('.json'):
        with open(file_path, 'w') as file:
            json.dump(data, file, indent=4)
    elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
        with open(file_path, 'w') as file:
            yaml.dump(data, file)
    else:
        raise ValueError("Unsupported file format. Use .json or .yaml/.yml")

def load_instance_from_file(instance, file_path):
    if file_path.endswith('.json'):
        with open(file_path, 'r') as file:
            data = json.load(file)
    elif file_path.endswith('.yaml') or file_path.endswith('.yml'):
        with open(file_path, 'r') as file:
            data = yaml.load(file, Loader=yaml.SafeLoader)
    else:
        raise ValueError("Unsupported file format. Use .json or .yaml/.yml")

    for key, value in data.items():
        setattr(instance, key, value)
        
        
        
        
if __name__ =="__main__":
        
    # Beispielklasse
    class MyClass:
        def __init__(self, name, age, occupation):
            self.name = name
            self.age = age
            self.occupation = occupation
            self.additional_info = {"hobbies": ["reading", "hiking"], "city": "Berlin"}
    
    # Beispielinstanz erstellen
    obj = MyClass("Alice", 30, "Engineer")
    
    # Speichern der Instanz in eine JSON-Datei
    save_instance_to_file(obj, 'instance.yaml')
    
    # Neue Instanz erstellen und Werte aus der Datei laden
    new_obj = MyClass("", 0, "")
    load_instance_from_file(new_obj, 'instance.yaml')
    print(new_obj.__dict__)
